"""This module contains a class for encoding spatial layers into a Matrix.

The 'LayerEncoder' class uses a shapegrid to generate a base matrix structure
and then each layer is encoded as a new column (or columns) for the resulting
encoded matrix.

Todo:
    Consider if we want to support hexagonal cells, if so, we will need to
        mask the resulting data windows and possibly change the minimum
        coverage calculation.

Note:
    Data array is oriented at top left (min x, max y)
"""
import json
import os
from random import shuffle

import numpy as np
from osgeo import gdal, ogr

from lmpy import Matrix

# DEFAULT_SCALE is the scale of the layer data array to the shapegrid cellsize
#     The number of data array cells in a (square) shapegrid cell is::
#         1.0 / DEFAULT_SCALE^2
DEFAULT_SCALE = 0.1  # Default scale for data array versus shapegrid cells
# DEFAULT_NODATA is the default value to use when no data is present in a
#    raster cell.
DEFAULT_NODATA = -9999


# .............................................................................
# Encoding methods


# .............................................................................
def _get_presence_absence_method(min_presence, max_presence, min_coverage, nodata):
    """Gets the function for determining presence for a data window.

    Args:
        min_presence (numeric): Data cells must have a value greater than or equal to
            this to be considered present.
        max_presence (numeric): Data cells must have a value less than or equal to this
            to be considered present.
        min_coverage (numeric): At least the percentage of the window must be
            classified as present to consider the window present.
        nodata (numeric): This values should be considered nodata.

    Returns:
        Method: A function for getting presence absence of a window.
    """
    if min_coverage >= 1.0:
        min_coverage = min_coverage / 100.0

    # ...............................
    def get_presence_absence(window):
        if window is None:
            return False
        min_num = max(min_coverage * window.size, 1)
        valid_cells = np.logical_and(
            np.logical_and(window >= min_presence, window <= max_presence),
            window != nodata)
        return np.sum(valid_cells) >= min_num

    return get_presence_absence


# .............................................................................
def _get_mean_value_method(nodata):
    """Gets the function to use for determining the mean value of a data window.

    Args:
        nodata (numeric): This value is assumed to be nodata in the array.

    Returns:
        Method: A function for getting the mean value of a window.
    """
    # ...............................
    def get_mean(window):
        if window is None:
            return nodata
        window_mean = np.nanmean(window)
        if np.isnan(window_mean):
            return nodata

        return window_mean

    return get_mean


# .............................................................................
def _get_largest_class_method(min_coverage, nodata):
    """Gets the function to use for determining the largest class.

    Args:
        min_coverage (numeric): The minimum percentage of the data window that must be
            covered by the largest class.
        nodata (numeric): This value is assumed to be nodata in the array

    Returns:
        Method: A function for getting the largest class in a window.
    """
    if min_coverage >= 1.0:
        min_coverage = min_coverage / 100.0

    # ...............................
    def get_largest_class(window):
        """Get largest class for numpy > 1.8.

        Args:
            window (Matrix): A window of layer data.

        Returns:
            ndarray: An array of encoded values.
        """
        if window is None:
            return nodata
        min_num = min_coverage * window.size
        largest_count = 0
        largest_class = nodata
        unique_values = np.column_stack(np.unique(window, return_counts=True))
        for class_, num in unique_values:
            if not np.isclose(class_, nodata) and num > largest_count \
                    and num > min_num:
                largest_class = class_
        return largest_class

    return get_largest_class


# .............................................................................
def _get_encode_hypothesis_method(hypothesis_values, min_coverage, nodata):
    """Gets the function to determine the hypothesis value for each data window.

    Args:
        hypothesis_values (list): A list of possible hypothesis values to look for.
            Each item in the list will be treated as its own column.
        min_coverage (numeric): The minimum percentage of each data window that must be
            covered by the returned hypothesis value.
        nodata (numeric): This value is assumed to be nodata

    Returns:
        Method: A function to encode the values in a window.
    """
    # Build the map
    val_map = {}
    i = 0
    for val in hypothesis_values:
        contrast_values = [-1, 1]
        shuffle(contrast_values)
        try:
            # Pair of values
            val_map[val[0]] = {
                'val': contrast_values[0],
                'index': i
            }
            val_map[val[1]] = {
                'val': contrast_values[1],
                'index': i
            }
        except Exception:
            # Single value
            val_map[val] = {
                'val': contrast_values[0],
                'index': i
            }
        i += 1
        # Note: 'i' is the number of values, we'll use that later

    if min_coverage >= 1.0:
        min_coverage = min_coverage / 100.0

    # ...............................
    def encode_method(window):
        """Encode method for numpy > 1.8.

        Args:
            window (Matrix): A window of layer data.

        Returns:
            ndarray: An array of encoded values.
        """
        if window is None:  # pragma: no cover
            return nodata
        min_vals = int(min_coverage * window.size)
        # Set default min count to min_vals
        # Note: This will cause last one to win if they are equal, change to
        #     '>' below and set this to * (min_vals - 1) to have first one win
        counts = np.zeros((i), dtype=int) * min_vals
        ret = np.zeros(i)

        unique_values = np.column_stack(np.unique(window, return_counts=True))

        # Check unique values in window
        for val, num in unique_values:
            if not np.isclose(val, nodata) and val in list(val_map.keys()) and\
                    num >= counts[val_map[val]['index']]:
                counts[val_map[val]['index']] = num
                ret[val_map[val]['index']] = val_map[val]['val']
        return ret

    return encode_method


# .............................................................................
class LayerEncoder:
    """The LayerEncoder class encodes layers into matrix columns.

    Attributes:
        encoded_matrix: A Matrix object with encoded layers.
    """

    # ...............................
    def __init__(self, shapegrid_filename):
        """Constructor for the layer encoder.

        Args:
            shapegrid_filename (str): A file path for the shapegrid.
        """
        # Process shapegrid
        self.shapegrid_filename = shapegrid_filename
        self._read_shapegrid(shapegrid_filename)

        self.encoded_matrix = None

    # ...............................
    def _encode_layer(self, window_func, encode_func, column_name, num_columns=1):
        """Encodes the layer using the provided encoding function.

        Args:
            window_func: A function that returns a window of array data for a
                provided x, y pair.
            encode_func: A function that encodes a window of array data.
            column_name: The header name to use for the column in the encoded
                matrix.
            num_columns: The number of columns that will be encoded by
                'encode_func'.  This can be non-zero if we are testing for
                multiple biogeographic hypotheses in a single vector layer for
                example.

        Returns:
            list: A list of column headers for the newly encoded columns.
        """
        shapegrid_dataset = ogr.Open(self.shapegrid_filename)
        shapegrid_layer = shapegrid_dataset.GetLayer()

        encoded_column = np.zeros((self.num_cells, num_columns))
        if num_columns == 1:
            column_headers = [column_name]
        else:  # pragma: no cover
            column_headers = [
                '{}-{}'.format(column_name, val) for val in range(num_columns)]

        row_headers = []

        i = 0

        feat = shapegrid_layer.GetNextFeature()
        while feat is not None:
            geom = feat.GetGeometryRef()
            cent = geom.Centroid()
            x_coord = cent.GetX()
            y_coord = cent.GetY()
            val = encode_func(window_func(x_coord, y_coord))
            encoded_column[i] = val
            row_headers.append((feat.GetFID(), x_coord, y_coord))
            i += 1
            feat = shapegrid_layer.GetNextFeature()

        col = Matrix(encoded_column,
                     headers={'0': row_headers,
                              '1': column_headers})

        if self.encoded_matrix is None:
            self.encoded_matrix = col
        else:
            self.encoded_matrix = Matrix.concatenate(
                [self.encoded_matrix, col], axis=1)
        # Return column headers for added columns
        return column_headers

    # ...............................
    @staticmethod
    def _get_window_function(data, layer_bbox, cell_size, num_cell_sides=4):
        """Gets a windowing function for the data.

        This function generates a function that will return a "window" of array
        data for a given (x, y) pair.

        Args:
            data: A numpy array with data for a layer
            layer_bbox: The bounding box of the layer in the map units of the
                layer.
            cell_size: Either a single value or a tuple with two values. If it
                is a single value, it will be used for both x and y cell sizes.
                If a tuple is provided, the first value will be used for the
                size of each cell in the x dimension and the second will be
                used for the size of the cell in the y dimension.
            num_cell_sides: The number of sides each shapegrid cell has::
                4 -- square
                6 -- hexagon

        Note:
            The origin (0, 0) of the data array should represent (min x, max y)
                for the layer.

        Raises:
            NotImplementedError: Raised if cell sides does note equal 4.

        Returns:
            Method: A function for processing a window of data.

        Todo:
            CJ - Enable hexagonal windows by masking data.
        """
        if num_cell_sides != 4:
            raise NotImplementedError('Only rectangular cells are currently supported')
        # Compute bounds here to save compute time
        y_size, x_size = data.shape
        min_x, min_y, max_x, max_y = layer_bbox

        # x_res = float(max_x - min_x) / x_size

        y_range = max_y - min_y
        x_range = max_x - min_x

        try:
            # Tuple or list
            x_size_2 = cell_size[0] / 2.0
            y_size_2 = cell_size[1] / 2.0
        except TypeError:  # pragma: no cover
            # Single value
            x_size_2 = y_size_2 = cell_size / 2.0

        # ...............................
        def get_rc(x_coord, y_coord):
            """Get the row and column values for a x and y coordinate pair.

            Args:
                x_coord (numeric): The x coordinate for a position in the layer.
                y_coord (numeric): The y coordinate for a postiion in the layer.

            Returns:
                tuple of int, int: Row and column values for the specified point.
            """
            x_prop = (1.0 * x_coord - min_x) / x_range
            y_prop = (1.0 * y_coord - min_y) / y_range

            col = int(x_size * x_prop)
            row = y_size - int(y_size * y_prop)
            return row, col

        # ...............................
        def window_function(x_coord, y_coord):
            """Get the array window from the centroid coordinates.

            Args:
                x_coord (numeric): The x coordinate for a position in the layer.
                y_coord (numeric): The y coordinate for a postiion in the layer.

            Returns:
                numeric: The value of the layer at the specified position.
            """
            # Note: Again, 0 row corresponds to top of map, so bigger y
            #     corresponds to lower row number
            # Upper left coorner
            uly, ulx = get_rc(x_coord - x_size_2, y_coord + y_size_2)
            # Lower right corner
            lry, lrx = get_rc(x_coord + x_size_2, y_coord - y_size_2)

            # If entire cell window is outside of layer, return None
            if max([uly, lry]) < 0 or max([ulx, lrx]) < 0 or \
                    min([uly, lry]) > y_size or min([ulx, lrx]) > x_size:
                return None

            return data[max(0, uly):min(y_size, lry),
                        max(0, ulx):min(x_size, lrx)]

        return window_function

    # ...............................
    def _read_layer(
        self,
        layer_filename,
        resolution=None,
        bbox=None,
        nodata=DEFAULT_NODATA,
        event_field=None
    ):
        """Reads a layer for processing.

        Args:
            layer_filename (str): The file path for the layer to read.
            resolution (numeric): An optional resolution to use for the input data if
                it is a vector layer.
            bbox (tuple): An optional bounding box in the form
                (min x, min y, max x, max y) to use if the layer is a vector layer.
            nodata (numeric): An optional nodata value to use if the layer is a vector
                layer.
            event_field (str): If provided, use this field as the burn value for a
                vector layer.

        Returns:
            tuple: A tuple containing a window function for returning a portion of the
                numpy array generated by the layer and the NODATA value to use with
                this layer.
        """
        # Get the file extension for the layer file name
        ext = os.path.splitext(layer_filename)[1]

        if ext == '.shp':
            window_func, nodata_value, events = self._read_vector_layer(
                layer_filename, resolution=resolution, bbox=bbox,
                nodata=nodata, event_field=event_field)
        else:
            window_func, nodata_value = self._read_raster_layer(layer_filename)
            events = set()

        return (window_func, nodata_value, events)

    # ...............................
    def _read_raster_layer(self, raster_filename):
        """Reads a raster layer for processing.

        Args:
            raster_filename: The file path for the raster layer.

        Returns:
            tuple: A tuple containing a window function for returning a portion of the
                numpy array generated by the layer and the NODATA value to use with
                this layer.
        """
        dataset = gdal.Open(raster_filename)
        band = dataset.GetRasterBand(1)
        layer_array = band.ReadAsArray().astype(float)
        nodata = band.GetNoDataValue()
        layer_array[np.where(layer_array == nodata)] = np.nan

        num_y, num_x = layer_array.shape
        min_x, x_res, _, max_y, _, y_res = dataset.GetGeoTransform()
        max_x = min_x + (num_x * x_res)
        min_y = max_y + (y_res * num_y)
        layer_bbox = (min_x, min_y, max_x, max_y)
        window_func = self._get_window_function(
            layer_array, layer_bbox, self.shapegrid_resolution,
            num_cell_sides=self.shapegrid_sides)

        return (window_func, nodata)

    # ...............................
    def _read_shapegrid(self, shapegrid_filename):
        """Read the shapegrid.

        Args:
            shapegrid_filename: The file location of the shapegrid.
        """
        shapegrid_dataset = ogr.Open(shapegrid_filename)
        self.shapegrid_layer = shapegrid_dataset.GetLayer()
        tmp = self.shapegrid_layer.GetExtent()
        self.shapegrid_bbox = (tmp[0], tmp[2], tmp[1], tmp[3])

        self.num_cells = self.shapegrid_layer.GetFeatureCount()

        feature_0 = self.shapegrid_layer.GetFeature(0)
        geom = feature_0.GetGeometryRef()

        # Get resolution and number of sides
        geom_wkt = geom.ExportToWkt()
        boundary_points = geom_wkt.split(',')
        if len(boundary_points) == 5:
            # Square
            envelope = geom.GetEnvelope()
            self.shapegrid_resolution = (envelope[1] - envelope[0],
                                         envelope[3] - envelope[2])
        else:  # pragma: no cover
            # TODO: Write tests for this when we want full support
            # Hexagon
            center = geom.Centroid()
            x_cent = center.GetX()
            y_cent = center.GetY()
            x_1, y_1 = boundary_points[1].split(' ')
            self.shapegrid_resolution = np.sqrt(
                (x_cent - x_1) ** 2 + (y_cent - y_1) ** 2)
        self.shapegrid_sides = len(boundary_points) - 1
        # self.shapegrid_layer.ResetReading()
        self.shapegrid_layer = None

    # ...............................
    def _read_vector_layer(
        self,
        vector_filename,
        resolution=None,
        bbox=None,
        nodata=DEFAULT_NODATA,
        event_field=None
    ):
        """Reads a vector layer for processing.

        Args:
            vector_filename: The vectorfile path for the layer to read.
            resolution: An optional resolution to use for the generated data
                array for the layer.
            bbox: An optional bounding box in the form
                (min x, min y, max x, max y) to use for the vector layer.  Will
                use the shapegrid bounding box if not provided.
            nodata: An optional nodata value to use if the layer is a vector
                layer.
            event_field: An optional shapefile attribute to use as the burn
                value for each cell.  This should be numeric.

        Returns:
            tuple: A tuple containing a window function for returning a portion of the
                numpy array generated by the layer, the NODATA value to use with this
                layer, and a set of distinct events to be used for processing.
        """
        options = ['ALL_TOUCHED=TRUE']
        if event_field is not None:
            options.append('ATTRIBUTE={}'.format(event_field))

        if resolution is None:
            resolution = [DEFAULT_SCALE * i for i in self.shapegrid_resolution]
        try:
            # Tuple or list
            x_res = resolution[0]
            y_res = resolution[1]
        except TypeError:  # pragma: no cover
            # Single value
            x_res = y_res = resolution

        if bbox is None:
            bbox = self.shapegrid_bbox

        min_x, min_y, max_x, max_y = bbox
        x_size = int(float(max_x - min_x) / x_res)
        y_size = int(float(max_y - min_y) / y_res)

        vector_ds = ogr.Open(vector_filename)
        vector_layer = vector_ds.GetLayer()

        raster_drv = gdal.GetDriverByName('MEM')
        raster_ds = raster_drv.Create(
            'temp', x_size, y_size, 1, gdal.GDT_Float32)
        raster_ds.SetGeoTransform((min_x, x_res, 0, max_y, 0, -1.0 * y_res))
        band = raster_ds.GetRasterBand(1)
        band.SetNoDataValue(nodata)
        band.FlushCache()
        init_ary = np.empty((y_size, x_size))
        init_ary.fill(nodata)
        band.WriteArray(init_ary)
        gdal.RasterizeLayer(raster_ds, [1], vector_layer, options=options)

        layer_array = raster_ds.ReadAsArray()
        raster_ds = None

        layer_bbox = (min_x, min_y, max_x, max_y)
        window_func = self._get_window_function(
            layer_array, layer_bbox, self.shapegrid_resolution,
            num_cell_sides=self.shapegrid_sides)

        distinct_events = list(np.unique(layer_array))
        try:
            # Go through list backwards to safely pop if needed
            for i in range(len(distinct_events) - 1, -1, -1):
                if np.isclose(distinct_events[i], nodata):
                    distinct_events.pop(i)
        except TypeError:  # pragma: no cover
            # This happens if only one value
            if np.isclose(distinct_events, nodata):
                distinct_events = []
            else:
                distinct_events = [distinct_events]
        return (window_func, nodata, distinct_events)

    # ...............................
    def encode_biogeographic_hypothesis(
        self,
        layer_filename,
        column_name,
        min_coverage,
        resolution=None,
        bbox=None,
        nodata=DEFAULT_NODATA,
        event_field=None
    ):
        """Encodes a biogeographic hypothesis layer.

        Encodes a biogeographic hypothesis layer by creating a Helmert contrast
        column in the encoded matrix.

        Args:
            layer_filename: The file location of the layer to encode.
            column_name: What to name this column in the encoded matrix.
            min_coverage: The minimum percentage of each data window that must
                be covered.
            resolution: If the layer is a vector, optionally use this as the
                resolution of the data grid.
            bbox: If the layer is a vector, optionally use this bounding box
                for the data grid.
            nodata: If the layer is a vector, optionally use this as the data
                grid nodata value.
            event_field: If the layer is a vector and contains multiple
                hypotheses, use this field to separate the vector file.

        Returns:
            list of str: A list of column headers for the newly encoded columns.
        """
        window_func, nodata, distinct_events = self._read_layer(
            layer_filename, resolution=resolution, bbox=bbox, nodata=nodata,
            event_field=event_field)
        if len(distinct_events) == 2:
            # Set the events to be opposite sides of same hypothesis
            distinct_events = [tuple(distinct_events)]
        encode_func = _get_encode_hypothesis_method(
            distinct_events, min_coverage, nodata)
        return self._encode_layer(
            window_func, encode_func, column_name,
            num_columns=len(distinct_events))

    # ...............................
    def encode_presence_absence(
        self,
        layer_filename,
        column_name,
        min_presence,
        max_presence,
        min_coverage,
        resolution=None,
        bbox=None,
        nodata=DEFAULT_NODATA,
        attribute_name=None
    ):
        """Encodes a distribution layer into a presence absence column.

        Args:
            layer_filename: The file location of the layer to encode.
            column_name: What to name this column in the encoded matrix.
            min_presence: The minimum value that should be treated as presence.
            max_presence: The maximum value to be considered as present.
            min_coverage: The minimum percentage of each data window that must
                be present to consider that cell present.
            resolution: If the layer is a vector, optionally use this as the
                resolution of the data grid.
            bbox: If the layer is a vector, optionally use this bounding box
                for the data grid.
            nodata: If the layer is a vector, optionally use this as the data
                grid nodata value.
            attribute_name: If the layer is a vector, use this field to
                determine presence.

        Returns:
            list of str: A list of column headers for the newly encoded columns.
        """
        window_func, nodata, _ = self._read_layer(
            layer_filename, resolution=resolution, bbox=bbox, nodata=nodata,
            event_field=attribute_name)
        encode_func = _get_presence_absence_method(
            min_presence, max_presence, min_coverage, nodata)
        return self._encode_layer(window_func, encode_func, column_name)

    # ...............................
    def encode_mean_value(
        self,
        layer_filename,
        column_name,
        resolution=None,
        bbox=None,
        nodata=DEFAULT_NODATA,
        attribute_name=None
    ):
        """Encodes a layer based on the mean value for each data window.

        Args:
            layer_filename: The file location of the layer to encode.
            column_name: What to name this column in the encoded matrix.
            resolution: If the layer is a vector, optionally use this as the
                resolution of the data grid.
            bbox: If the layer is a vector, optionally use this bounding box
                for the data grid.
            nodata: If the layer is a vector, optionally use this as the data
                grid nodata value.
            attribute_name: If the layer is a vector, use this field to
                determine value.

        Returns:
            list of str: A list of column headers for the newly encoded columns
        """
        window_func, nodata, _ = self._read_layer(
            layer_filename, resolution=resolution, bbox=bbox, nodata=nodata,
            event_field=attribute_name)
        encode_func = _get_mean_value_method(nodata)
        return self._encode_layer(window_func, encode_func, column_name)

    # ...............................
    def encode_largest_class(
        self,
        layer_filename,
        column_name,
        min_coverage,
        resolution=None,
        bbox=None,
        nodata=DEFAULT_NODATA,
        attribute_name=None
    ):
        """Encodes a layer based on the largest class in each data window.

        Args:
            layer_filename: The file location of the layer to encode.
            column_name: What to name this column in the encoded matrix.
            min_coverage: The minimum percentage of each data window that must
                be the covered by the largest class.
            resolution: If the layer is a vector, optionally use this as the
                resolution of the data grid.
            bbox: If the layer is a vector, optionally use this bounding box
                for the data grid.
            nodata: If the layer is a vector, optionally use this as the data
                grid nodata value.
            attribute_name: If the layer is a vector, use this field to
                determine largest class.

        Returns:
            list of str: A list of column headers for the newly encoded columns.
        """
        window_func, nodata, _ = self._read_layer(
            layer_filename,
            resolution=resolution,
            bbox=bbox,
            nodata=nodata,
            event_field=attribute_name
        )
        encode_func = _get_largest_class_method(min_coverage, nodata)
        return self._encode_layer(window_func, encode_func, column_name)

    # ...............................
    def get_encoded_matrix(self):
        """Returns the encoded matrix.

        Returns:
            Matrix: The encoded matrix as a Matrix object
        """
        return self.encoded_matrix

    # ...............................
    def get_geojson(self):
        """Formats the encoded matrix as GeoJSON.

        Returns:
            dict: A JSON dictionary for the encoded matrix.
        """
        ret = {
            'type': 'FeatureCollection'
        }
        features = []

        column_headers = self.encoded_matrix.get_column_headers()

        column_enum = [(j, str(k)) for j, k in enumerate(column_headers)]

        shapegrid_dataset = ogr.Open(self.shapegrid_filename)
        shapegrid_layer = shapegrid_dataset.GetLayer()

        i = 0
        feat = shapegrid_layer.GetNextFeature()
        while feat is not None:
            ft_json = json.loads(feat.ExportToJson())
            # right_hand_rule(ft_json['geometry']['coordinates'])
            # TODO(CJ): Remove this if updated library adds first id correctly
            ft_json['id'] = feat.GetFID()
            ft_json['properties'] = {
                k: self.encoded_matrix[i, j].item() for j, k in column_enum}
            features.append(ft_json)
            i += 1
            feat = shapegrid_layer.GetNextFeature()

        ret['features'] = features
        shapegrid_dataset = shapegrid_layer = None
        return ret


# .............................................................................
__all__ = ['LayerEncoder']



"""Module containing a class for working with a spatial index.

Version 1: Store geometries in memory in table.  Save as wkt.
"""
import json
import os
from osgeo import ogr
import rtree


# .............................................................................
def create_geometry_from_bbox(min_x, min_y, max_x, max_y):
    """Create a geometry from a bounding box.

    Args:
        min_x (numeric): The minimum x value for the bounding box.
        min_y (numeric): The minimum y value for the bounding box.
        max_x (numeric): The maximum x value for the bounding box.
        max_y (numeric): The maximum y value for the bounding box.

    Returns:
        str: A well-known text representation of a bounding box geometry.
    """
    wkt = 'POLYGON (({0} {1}, {0} {3}, {2} {3}, {2} {1}, {0} {1}))'.format(
        min_x, min_y, max_x, max_y)
    return ogr.CreateGeometryFromWkt(wkt)


# .............................................................................
def quadtree_index(geom, bbox, min_size, depth_left):
    """Use a quadtree approach to gather spatial index data.

    Args:
        geom (ogr.Geometry): A geometry object to index.
        bbox (tuple): The bounding box to intersect with the geometry.
        min_size (float): The minimum intersection size to count as intersecting.
        depth_left (int): The number of divisions left for intersecting smaller
            bounding boxes with portions of the geometry.

    Returns:
        list of tuple: A list of bounding box, geometry or true tuples used to create a
            quadtree index of a geometry.
    """
    # min_x, min_y, max_x, max_y = bbox
    test_geom = create_geometry_from_bbox(*bbox)
    intersection = geom.Intersection(test_geom)
    min_x, max_x, min_y, max_y = intersection.GetEnvelope()
    if min_x == max_x or min_y == max_y:
        return []
    # if intersection.Equals(test_geom):
    if intersection.Area() < min_size:
        return [(bbox, intersection)]
    if intersection.Area() == test_geom.Area():
        return [(bbox, True)]
    ret = []
    if depth_left > 0:
        half_x = min_x + (max_x - min_x) / 2.0
        half_y = min_y + (max_y - min_y) / 2.0
        ret.extend(
            quadtree_index(
                intersection, (min_x, min_y, half_x, half_y), min_size,
                depth_left - 1))
        ret.extend(
            quadtree_index(
                intersection, (half_x, min_y, max_x, half_y), min_size,
                depth_left - 1))
        ret.extend(
            quadtree_index(
                intersection, (half_x, half_y, max_x, max_y), min_size,
                depth_left - 1))
        ret.extend(
            quadtree_index(
                intersection, (min_x, half_y, half_x, max_y), min_size,
                depth_left - 1))
    return ret


# .............................................................................
class SpatialIndex:
    """This class provides an index for quickly performing intersects."""
    # ..........................
    def __init__(self, index_name=None):
        """Constructor for the spatial index.

        Args:
            index_name (str): A name to use for saving the index to a file.
        """
        self.index = rtree.index.Index(index_name)
        self._att_filename = '{}.json'.format(index_name)
        self._geom_filename = '{}.geom_json'.format(index_name)
        self.att_lookup = {}
        if os.path.exists(self._att_filename):
            with open(self._att_filename) as in_file:
                self.att_lookup = json.load(in_file)
        self.geom_lookup = {}
        if os.path.exists(self._geom_filename):
            with open(self._geom_filename) as in_file:
                tmp_geoms = json.load(in_file)
                for k, wkt in tmp_geoms.items():
                    self.geom_lookup[str(k)] = ogr.CreateGeometryFromWkt(wkt)
                # self.geom_lookup = json.load(in_file)
        self.min_size = 0.01
        self.depth_left = 10
        self.next_geom = len(self.geom_lookup)

    # ..........................
    def add_feature(self, identifier, geom, att_dict):
        """Add a feature to the index.

        Args:
            identifier (str): An identifier for this feature in the lookup table.
            geom (ogr.Geometry): A geometry to spatially index.
            att_dict (dict): A dictionary of attributes to store in the lookup table.
        """
        try:
            self.att_lookup[str(identifier)] = att_dict
            if isinstance(geom, str):
                geom = ogr.CreateGeometryFromWkt(geom)
            min_x, max_x, min_y, max_y = geom.GetEnvelope()
            idx_entries = quadtree_index(
                geom, (min_x, min_y, max_x, max_y),
                self.min_size, self.depth_left)
            # print('{} - Num idx entries: {}'.format(
            #     identifier, len(idx_entries)))
            for bbox, idx_geom in idx_entries:
                if isinstance(idx_geom, bool) and idx_geom:
                    # Index as entire bbox
                    self.index.insert(identifier, bbox, obj=True)
                else:
                    # Add geometry to lookup, increment counter
                    self.index.insert(identifier, bbox, obj=self.next_geom)
                    self.geom_lookup[str(self.next_geom)] = idx_geom
                    self.next_geom += 1
        except Exception as err:  # pragma: no cover
            print(err)
            print(identifier)
            print(att_dict)

    # ..........................
    def close(self):
        """Close the index."""
        self.index.close()

    # ..........................
    def save(self):
        """Save the index attributes."""
        with open(self._att_filename, 'w') as out_file:
            json.dump(self.att_lookup, out_file)
        with open(self._geom_filename, 'w') as out_file:
            out_geoms = {
                k: val.ExportToWkt() for k, val in self.geom_lookup.items()}
            json.dump(out_geoms, out_file)
            # json.dump(self.geom_lookup, out_file)

    # ..........................
    def search(self, x, y):
        """Search for x, y and return attributes in lookup if found.

        Args:
            x (numeric): The x coordinate to search for.
            y (numeric): The y coordinate to search for.

        Returns:
            dict: A dictionary of index hits for the search.
        """
        hits = {}
        for hit in self.index.intersection((x, y, x, y), objects=True):
            if hit.id not in hits.keys():
                if isinstance(hit.object, bool) or \
                    self._point_intersect(
                        x, y, self.geom_lookup[str(hit.object)]):
                    hits[str(hit.id)] = self.att_lookup[str(hit.id)]
        return hits

    # ..........................
    @staticmethod
    def _point_intersect(pt_x, pt_y, geom):
        """Perform a point intersect.

        Args:
            pt_x (numeric): The x coordinate of the point.
            pt_y (numeric): The y coordinate of the point.
            geom (ogr.Geometry): The geometry to intersect.

        Returns:
            bool: Indication if the point is within the geometry.
        """
        pt_geom = ogr.CreateGeometryFromWkt('POINT ({} {})'.format(pt_x, pt_y))
        return pt_geom.Within(geom)


# .............................................................................
__all__ = ['SpatialIndex']



"""Encode a tree into matrix for faster statistics computations."""
import argparse

import numpy as np

from lmpy import Matrix, TreeWrapper


# .....................................................................................
def encode_tree(tree):
    """Encode a tree into a binary matrix and two data arrays for node and tips.

    Args:
        tree (TreeWrapper): A tree object to encode.

    Returns:
        (Matrix, Matrix, Matrix): The encoded tree matrix, node heights, tip lengths.
    """
    num_tips = len(tree.taxon_namespace)
    num_nodes = len(tree.nodes()) - num_tips

    tip_lengths = Matrix(np.zeros((num_tips,)))
    node_heights = Matrix(np.zeros((num_nodes,)))

    ordered_taxa = [taxon.label for taxon in tree.taxon_namespace]
    node_labels = ['Node {}'.format(i) for i in range(num_nodes)]

    taxon_index_map = {
        val: idx for idx, val in enumerate(ordered_taxa)
    }

    tree_mtx = Matrix(
        np.zeros((num_tips, num_nodes), dtype=np.bool),
        headers={'0': ordered_taxa, '1': node_labels}
    )

    def process_node(node, node_idx):
        node_taxa = []
        node_height = 0
        next_node_idx = node_idx + 1
        for child in node.child_nodes():
            if child.is_leaf():
                if child.edge_length < 1e-05:
                    print(
                        'Short branch length {:.6f} child.edge_length)'.format(
                            child.edge_length
                        )
                    )
                try:
                    tax_index = taxon_index_map[child.taxon.label]
                    node_taxa.append(tax_index)
                    tip_lengths[tax_index] = child.edge_length
                    tree_mtx[tax_index, node_idx] = 1
                    node_height = child.edge_length
                except Exception as err:
                    print(err)
                    print(child.__dict__)
                    print(child.annotations)
                    raise err
            else:
                child_taxa, next_node_idx, child_node_height = process_node(
                    child, next_node_idx
                )
                node_height = child_node_height + child.edge_length

                node_taxa.extend(child_taxa)

                for child_tax in child_taxa:
                    tree_mtx[child_tax, node_idx] = 1

        node_heights[node_idx] = node_height

        return node_taxa, next_node_idx, node_height

    process_node(tree.seed_node, 0)

    return tree_mtx, node_heights, tip_lengths


# .....................................................................................
def main():
    """Main controlling method for script."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'tree_filename',
        type=str,
        help='File path to a phylogenetic tree.'
    )
    parser.add_argument(
        'tree_schema',
        type=str,
        help='The format of the tree file (ex. newick or nexus).'
    )
    parser.add_argument(
        'out_tree_matrix_filename',
        type=str,
        help='File path to write encoded tree matrix (tip rows by node columns).'
    )
    parser.add_argument(
        'out_node_heights_filename',
        type=str,
        help='File path to write node heights matrix.'
    )
    parser.add_argument(
        'out_tip_lengths_filename',
        type=str,
        help='File path to write tip lengths matrix.'
    )
    args = parser.parse_args()

    tree = TreeWrapper.get(path=args.tree_filename, schema=args.tree_schema)
    tree_mtx, node_heights, tip_lengths = encode_tree(tree)
    tree_mtx.write(args.out_tree_matrix_filename)
    node_heights.write(args.out_node_heights_filename)
    tip_lengths.write(args.out_tip_lengths_filename)


# .....................................................................................
if __name__ == '__main__':
    main()



"""Script to encode layers into a Matrix."""
import argparse

from lmpy.data_preparation.layer_encoder import LayerEncoder


DESCRIPTION = 'Encode raster and / or vector layers into a site by layer matrix.'


# .....................................................................................
def cli():
    """Command line interface for layer encoding.

    Raises:
        ValueError: Raised if an unknown encoding method is provided.
    """
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '--encode_method',
        '-m',
        type=str,
        choices=['biogeo', 'presence_absence', 'largest_class', 'mean_value'],
        default='mean_value',
        help='The method to use to encode each layer.'
    )
    parser.add_argument(
        '--min_coverage',
        type=float,
        default=25,
        help='Minimum percentage of a cell that has to be covered to encode it.'
    )
    parser.add_argument(
        '--min_presence',
        type=float,
        default=10,
        help='Minimum value to be considered present when encoding presence absence.'
    )
    parser.add_argument(
        '--max_presence',
        type=float,
        default=127,
        help='The maximum value that should be considered present.'
    )
    parser.add_argument(
        '--attribute_field',
        type=str,
        help='The vector attribute to use for encoding layers.'
    )
    parser.add_argument(
        'shapegrid_filename',
        type=str,
        help='File location of shapegrid to use for site geometries.'
    )
    parser.add_argument(
        'out_matrix_filename',
        type=str,
        help='File location to write the encoded matrix.'
    )
    # Layers to encode
    parser.add_argument(
        '--layer',
        '-l',
        nargs=2,
        action='append',
        help='File location of layer followed by a label.'
    )

    args = parser.parse_args()

    encoder = LayerEncoder(args.shapegrid_filename)

    for lyr_fn, lyr_label in args.layer:
        if args.encode_method == 'biogeo':
            encoder.encode_biogeographic_hypothesis(
                lyr_fn,
                lyr_label,
                args.min_coverage,
                event_field=args.attribute_field
            )
        elif args.encode_method == 'presence_absence':
            encoder.encode_presence_absence(
                lyr_fn,
                lyr_label,
                args.min_presence,
                args.max_presence,
                args.min_coverage,
                attribute_name=args.attribute_field
            )
        elif args.encode_method == 'largest_class':
            encoder.encode_largest_class(
                lyr_fn,
                lyr_label,
                args.min_coverage,
                attribute_name=args.attribute_field
            )
        elif args.encode_method == 'mean_value':
            encoder.encode_mean_value(
                lyr_fn,
                lyr_label,
                attribute_name=args.attribute_field
            )
        else:
            raise ValueError('Unknown encoding method: {}'.format(args.encode_method))

    enc_mtx = encoder.get_encoded_matrix()
    enc_mtx.write(args.out_matrix_filename)


# .....................................................................................
__all__ = ['cli']


# .....................................................................................
if __name__ == '__main__':
    cli()



"""This tool cleans occurrence records according to the wranglers specified."""
import argparse
import json

from lmpy.data_wrangling.occurrence.factory import wrangler_factory
from lmpy.point import PointCsvReader, PointCsvWriter


# .....................................................................................
def clean_data(reader, writer_filename, wranglers, write_fields=None):
    """Clean occurrence data.

    Args:
        reader (PointCsvReader): A reader object that generates Point objects.
        writer_filename (str): A file location to write cleaned points.
        wranglers (list of Wranglers): A list of data wranglers that manipulate
            and / or filter Points for cleaning.
        write_fields (list or None): A list of Point attributes to write to output CSV.
            If None, determine from first cleaned Point.
    """
    # Open reader
    reader.open()
    writer = None
    for points in reader:
        for wrangler in wranglers:
            # If there are points, wrangle them
            if points:
                points = wrangler(points)
        # If any points are left, write them
        if points:
            if writer is None:
                if write_fields is None:
                    write_fields = points[0].get_attribute_names()
                writer = PointCsvWriter(writer_filename, write_fields)
                writer.open()
            writer.write_points(points)
    # Close reader and writer
    reader.close()
    if writer:
        writer.close()


# .....................................................................................
def cli():
    """A command-line interface to the tool."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-sp', '--species_key', type=str, default='species_name',
        help='The CSV column header for species name.'
    )
    parser.add_argument(
        '-x', '--x_key', type=str, default='x',
        help='The CSV column header for the X (longitude) field.'
    )
    parser.add_argument(
        '-y', '--y_key', type=str, default='y',
        help='The CSV column header for the Y (latitude) field.'
    )
    parser.add_argument(
        'reader_filename', type=str, help='Input occurrence CSV filename.'
    )
    parser.add_argument(
        'writer_filename', type=str, help='Output cleaned occurrence CSV filename.'
    )
    parser.add_argument(
        'wrangler_config_filename', type=str, help='Wrangler(s) configuration file.'
    )
    args = parser.parse_args()

    # Get wranglers
    wranglers = wrangler_factory(
        json.load(open(args.wrangler_config_filename, mode='rt'))
    )

    # Get reader
    reader = PointCsvReader(
        args.reader_filename, args.species_key, args.x_key, args.y_key
    )

    # Clean data
    clean_data(reader, args.writer_filename, wranglers)


# .....................................................................................
__all__ = ['clean_data', 'cli']


# .....................................................................................
if __name__ == '__main__':
    cli()



"""Randomize PAMs using CJ's algorithm.

This module contains functions used to randomize a PAM using CJ's algorithm.
This algorithm can run in a parallel fashion and uses a fill-based approach so
as to prevent a bias caused by starting with an initial condition of a
populated matrix.
"""
import numpy as np

from lmpy.matrix import Matrix

# Number of times to look for a match when fixing problems
SEARCH_THRESHOLD = 100


# .............................................................................
def fill_shuffle_reshape_heuristic(orig_pam):
    """Create an approximation with the correct number of 1s randomly placed.

    Creates an array with the total number of ones from the original PAM and
    shuffles it then it reshapes it to match the shape of the original PAM.

    Args:
        orig_pam (Matrix): The observed PAM to randomize.

    Returns:
        Matrix: An approximate randomization of the PAM that is uncorrected.
    """
    fill = int(np.sum(orig_pam))
    approx = np.zeros((orig_pam.shape), dtype=int)
    approx[:fill] = 1
    np.random.shuffle(approx)
    return approx.reshape(orig_pam.shape)


# .............................................................................
def total_fill_percentage_heuristic(orig_pam):
    """Create an approximation using the total fill percentage of the PAM.

    Creates an approximation using the total matrix fill of the original PAM
    as a weight threshold to compare with randomly generated numbers.

    Args:
        orig_pam (:obj:`Matrix`): The observed PAM to randomize.

    Returns:
        Matrix: An approximate randomization of the PAM that is uncorrected.
    """
    fill = np.sum(orig_pam)
    fill_percentage = 1.0 * fill / orig_pam.size
    approx = (
        np.random.uniform(
            low=0.0, high=1.0, size=orig_pam.shape
        ) <= fill_percentage
    ).astype(int)
    return approx


# .............................................................................
def max_col_or_row_heuristic(orig_pam):
    """Weighting method using max weight between row and column.

    This method returns a matrix of weights where the weight of each cell is
    the maximum between the proportions of the row and col

    Args:
        orig_pam (:obj:`Matrix`): The observed PAM to randomize.

    Returns:
        Matrix: An approximate randomization of the PAM that is uncorrected.
    """
    row_totals = np.sum(orig_pam, axis=1)
    col_totals = np.sum(orig_pam, axis=0)

    row_weights = row_totals.astype(float) / row_totals.shape[0]
    col_weights = col_totals.astype(float) / col_totals.shape[0]

    row_weights = np.expand_dims(
        row_totals.astype(float) / row_totals.shape[0], 1)
    col_weights = np.expand_dims(
        col_totals.astype(float) / col_totals.shape[0], 0)
    return (
        np.random.uniform(
            low=0.0, high=1.0, size=orig_pam.shape
        ) <= np.maximum(row_weights, col_weights)
    ).astype(int)


# .............................................................................
def min_col_or_row_heuristic(orig_pam):
    """Weighting method using max weight between row and column.

    This method returns a matrix of weights where the weight of each cell is
    the maximum between the proportions of the row and col

    Args:
        orig_pam (:obj:`Matrix`): The observed PAM to randomize.

    Returns:
        Matrix: An approximate randomization of the PAM that is uncorrected.
    """
    row_totals = np.sum(orig_pam, axis=1, dtype=int)
    col_totals = np.sum(orig_pam, axis=0, dtype=int)

    row_weights = np.expand_dims(
        row_totals.astype(float) / row_totals.shape[0], 1)
    col_weights = np.expand_dims(
        col_totals.astype(float) / col_totals.shape[0], 0)
    return (
        np.random.uniform(
            low=0.0, high=1.0, size=orig_pam.shape
        ) <= np.minimum(row_weights, col_weights, dtype=np.single)
    ).astype(int)


# .............................................................................
def all_zeros_heuristic(orig_pam):
    """Creates a two-dimensional approximation composed of all zeros.

    Args:
        orig_pam (:obj:`Matrix`): The observed PAM to randomize.

    Returns:
        Matrix: An approximate randomization of the PAM that is uncorrected.
    """
    return np.zeros(orig_pam.shape, dtype=int)


# .............................................................................
def all_ones_heuristic(orig_pam):
    """Creates a two-dimensional approximation composed of all ones.

    Args:
        orig_pam (:obj:`Matrix`): The observed PAM to randomize.

    Returns:
        Matrix: An approximate randomization of the PAM that is uncorrected.
    """
    return np.ones(orig_pam.shape, dtype=int)


# .............................................................................
def grady_randomize(mtx, approximation_heuristic=total_fill_percentage_heuristic):
    """Main function for creating a random matrix.

    Args:
        mtx (:obj:`Matrix`): A Matrix object representation of a PAM
        approximation_heuristic (:obj:`Method`): A function that generates an
            approximation of a final randomized matrix.

    Returns:
        Matrix: A matrix of random presence absence values with the same
            marginal totals as the input matrix 'mtx'.
    """
    # Step 0. Get marginal totals
    # ...........................
    num_rows, num_cols = mtx.shape
    row_totals = np.sum(mtx, axis=1).reshape((num_rows, 1))
    col_totals = np.sum(mtx, axis=0).reshape((1, num_cols))
    valid_rows = np.sum(mtx, axis=1) != 0
    valid_cols = np.sum(mtx, axis=0) != 0

    # Step 1. Get Initial random matrix
    # ...........................
    # Get approximation
    rand_mtx_data = approximation_heuristic(mtx)

    # Step 2: Purge over-filled marginals
    # ...................................
    # For each row / column with more ones than the marginal total, remove
    #    extra ones until within limit
    row_sums = np.sum(rand_mtx_data, axis=1).reshape((num_rows, 1))
    for i in np.where(row_sums > row_totals)[0]:
        row_choices = np.where(rand_mtx_data[i] == 1)[0]
        change_count = int(row_sums[i] - row_totals[i])
        rand_mtx_data[
            i, np.random.permutation(row_choices)[:change_count]] = False

    col_sums = np.sum(rand_mtx_data, axis=0).reshape((1, num_cols))
    for j in np.where(col_sums > col_totals)[1]:
        col_choices = np.where(rand_mtx_data[:, j] == 1)[0]
        change_count = int(col_sums[0, j] - col_totals[0, j])
        rand_mtx_data[
            np.random.permutation(col_choices)[:change_count], j] = False

    # Step 3: Fill
    # ...........................
    problem_rows = []
    problem_cols = []

    row_sums = np.sum(rand_mtx_data, axis=1)
    col_sums = np.sum(rand_mtx_data, axis=0)

    unfilled_cols = np.where(col_sums < col_totals[0, :])[0].tolist()
    for row_idx in np.random.permutation(
            np.where(row_sums < row_totals[:, 0])[0]):

        possible_cols = np.random.permutation(
            np.intersect1d(
                np.where(rand_mtx_data[row_idx, :] == 0)[0],
                unfilled_cols, assume_unique=True))
        num_to_fill = int(row_totals[row_idx, 0] - row_sums[row_idx])
        if num_to_fill > possible_cols.shape[0]:
            # Add this row to problem rows because we can't fill it enough
            problem_rows.append(row_idx)
            # Just fill what we can
            num_to_fill = possible_cols.shape[0]
        # Fill cells
        rand_mtx_data[row_idx, possible_cols[:num_to_fill]] = True

        # Add to col_sums
        col_sums[possible_cols[:num_to_fill]] += 1

        # Check if we should remove these columns from the list
        for c in np.where(
            col_sums[possible_cols[:num_to_fill]] == col_totals[
                0, possible_cols[:num_to_fill]])[0]:
            unfilled_cols.remove(possible_cols[c])

    problem_cols = unfilled_cols

    # Step 4: Fix problems
    # ...........................
    j = 0
    while problem_rows:
        r = np.random.choice(problem_rows)
        c = np.random.choice(problem_cols)
        i = 0

        cs = np.where((rand_mtx_data[r] == 0) & (valid_cols))[0]
        rs = np.where((rand_mtx_data[:, c] == 0) & (valid_rows))[0]

        num_tries = 0
        found = False

        while not found and num_tries < SEARCH_THRESHOLD:
            r2 = np.random.choice(rs)
            c2 = np.random.choice(cs)
            num_tries += 1

            if rand_mtx_data[r2, c2] == 1:
                rand_mtx_data[r, c2] = 1
                rand_mtx_data[r2, c2] = 0
                rand_mtx_data[r2, c] = 1
                found = True

        if not found:  # pragma: no cover
            # Can't fix row easily so randomly move a one
            prs = np.where(rand_mtx_data[:, c2] == 1)[0]
            update_row = np.random.choice(prs)
            rand_mtx_data[r, c2] = 1
            rand_mtx_data[update_row, c2] = 0

            # Should we add update row to problem rows or is it already there?
            if update_row not in problem_rows:
                problem_rows.append(update_row)

        r_sum = np.sum(rand_mtx_data[r, :])
        c_sum = np.sum(rand_mtx_data[:, c])

        if r_sum == int(row_totals[r]):
            problem_rows.remove(r)
        if c_sum == int(col_totals[0, c]):
            problem_cols.remove(c)

    row_sums = np.sum(rand_mtx_data, axis=1)

    return Matrix(rand_mtx_data, headers=mtx.get_headers())


# .............................................................................
__all__ = [
    'all_ones_heuristic',
    'all_zeros_heuristic',
    'fill_shuffle_reshape_heuristic',
    'grady_randomize',
    'max_col_or_row_heuristic',
    'min_col_or_row_heuristic',
    'total_fill_percentage_heuristic'
]



"""This script will randomize a PAM while maintaining marginal totals."""
import argparse

from lmpy import Matrix
from lmpy.randomize.grady import grady_randomize


DESCRIPTION = '''\
Randomize a binary presence-absence matrix (PAM) while maintaining marginal totals.\
'''


# .....................................................................................
def randomize_pam(in_pam):
    """Randomize a PAM.

    Args:
        in_pam (Matrix): The input PAM to be randomized.

    Returns:
        Matrix: A randomized PAM that maintains marginal totals.
    """
    random_pam = grady_randomize(in_pam)
    return random_pam


# .....................................................................................
def cli():
    """Function providing a command line interface to the tool."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        'input_pam_filename',
        type=str,
        help='The file location of the PAM to randomize.'
    )
    parser.add_argument(
        'output_pam_filename',
        type=str,
        help='The file location to write the randomized PAM.'
    )
    args = parser.parse_args()
    in_pam = Matrix.load(args.input_pam_filename)
    rand_pam = randomize_pam(in_pam)
    rand_pam.write(args.output_pam_filename)


# .....................................................................................
__all__ = ['cli', 'randomize_pam']


# .....................................................................................
if __name__ == '__main__':
    cli()



"""Tools module."""
from . import clean_occurrences
from . import convert_csv_to_lmm
from . import convert_lmm_to_csv
from . import encode_layers
from . import randomize_pam

__all__ = []
__all__.extend(clean_occurrences.__all__)
__all__.extend(convert_csv_to_lmm.__all__)
__all__.extend(convert_lmm_to_csv.__all__)
__all__.extend(encode_layers.__all__)
__all__.extend(randomize_pam.__all__)



"""Convert a numeric (.csv) file to a lmpy Matrix (.lmm) file."""
import argparse

from lmpy import Matrix


DESCRIPTION = 'Convert a CSV file of numerical values into a lmpy Matrix.'


# .....................................................................................
def convert_csv_to_lmm(csv_filename, num_header_rows, num_header_cols):
    """Convert a CSV file to a lmpy lmm Matrix.

    Args:
        csv_filename (str): The file location of the csv to convert.
        num_header_rows (int): The number of header rows in the csv.
        num_header_cols (int): The number of header columns in the csv.

    Returns:
        Matrix: A matrix from the converted csv file.
    """
    with open(csv_filename, mode='rt') as csv_in:
        mtx = Matrix.load_csv(
            csv_in,
            num_header_rows=num_header_rows,
            num_header_cols=num_header_cols
        )
    return mtx


# .....................................................................................
def cli():
    """Provide a command-line tool for converting csvs to lmms."""
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        '--header_rows',
        '-r',
        type=int,
        default=0,
        help='The number of header rows in the CSV.'
    )
    parser.add_argument(
        '--header_cols',
        '-c',
        type=int,
        default=0,
        help='The number of header columns in the CSV.'
    )
    parser.add_argument(
        'in_csv_filename',
        type=str,
        help='CSV filename to convert to lmm Matrix.'
    )
    parser.add_argument('out_lmm_filename', type=str, help='Filename for .lmm Matrix.')
    args = parser.parse_args()
    mtx = convert_csv_to_lmm(args.in_csv_filename, args.header_rows, args.header_cols)
    mtx.write(args.out_lmm_filename)


# .....................................................................................
__all__ = ['cli', 'convert_csv_to_lmm']


# .....................................................................................
if __name__ == '__main__':
    cli()



"""Tests the running statistics module."""
import numpy as np
import pytest

from lmpy import Matrix
from lmpy.statistics.running_stats import (
    compare_absolute_values, compare_signed_values, RunningStats)


# .............................................................................
class Test_RunningStats(object):
    """Test the RunningStats class."""
    # .....................................
    def test_single_values(self):
        """Tests the RunningStats class using single values."""
        vals1 = [1, 2, 3, 4, 5]
        vals2 = [6, 7, 7, 8, 9, 10]
        all_vals = []
        all_vals.extend(vals1)
        all_vals.extend(vals2)
        np_all_vals = np.array(all_vals)

        rs = RunningStats()
        assert rs.variance == 0.0
        for v in vals1:
            rs.push(v)
        rs.push(vals2)
        assert rs.mean == np.mean(np_all_vals)
        assert rs.standard_deviation == np.std(np_all_vals, ddof=1)
        assert rs.variance == np.var(np_all_vals, ddof=1)
        with pytest.raises(Exception):
            rs.p_values

    # .....................................
    def test_single_values_with_p_values(self):
        """Tests the RunningStats class using single values with p-values."""
        f_val = 6
        vals1 = [1, 2, 3, 4, 5]
        vals2 = [6, -7, 7, -8, -9, 10]
        all_vals = []
        all_vals.extend(vals1)
        all_vals.extend(vals2)
        np_all_vals = np.array(all_vals)

        rs = RunningStats(observed=f_val, compare_fn=compare_absolute_values)
        assert rs.variance == 0.0
        for v in vals1:
            rs.push(v)
        rs.push(vals2)
        assert np.isclose(rs.mean, np.mean(np_all_vals))
        assert np.isclose(rs.standard_deviation, np.std(np_all_vals, ddof=1))
        assert np.isclose(rs.variance, np.var(np_all_vals, ddof=1))
        num_greater = 0
        for v in all_vals:
            if np.abs(v) > np.abs(f_val):
                num_greater += 1
        assert rs.p_values == float(num_greater) / len(all_vals)

    # .....................................
    def test_single_values_with_p_values_signed(self):
        """Tests the RunningStats class using single values with p-values."""
        f_val = 6
        vals1 = [1, 2, 3, 4, 5]
        vals2 = [6, -7, 7, -8, 9, 10]
        all_vals = []
        all_vals.extend(vals1)
        all_vals.extend(vals2)
        np_all_vals = np.array(all_vals)

        rs = RunningStats(observed=f_val, compare_fn=compare_signed_values)
        assert rs.variance == 0.0
        for v in vals1:
            rs.push(v)
        rs.push(vals2)
        assert np.isclose(rs.mean, np.mean(np_all_vals))
        assert np.isclose(rs.standard_deviation, np.std(np_all_vals, ddof=1))
        assert np.isclose(rs.variance, np.var(np_all_vals, ddof=1))
        num_greater = 0
        for v in all_vals:
            if v > f_val:
                num_greater += 1
        assert rs.p_values == float(num_greater) / len(all_vals)

    # .....................................
    def test_matrix_with_no_p_values(self):
        """Tests the RunningStats class using Matrix objects without p-values."""
        _ = Matrix(np.random.randint(0, 10, size=(10, 10)))
        vals = []
        rs = RunningStats()
        assert np.all(rs.variance == np.zeros((10, 10)))
        for _ in range(5):
            vals.append(Matrix(np.random.randint(-10, 10, size=(10, 10))))
        rs.push(vals)
        for _ in range(5):
            test_val = Matrix(np.random.randint(-10, 10, size=(10, 10)))
            rs.push(test_val)
            vals.append(test_val)
        v_stack = np.array([v for v in vals])
        assert np.all(np.isclose(rs.mean, np.mean(v_stack, axis=0)))
        assert np.all(
            np.isclose(
                rs.standard_deviation, np.std(v_stack, axis=0, ddof=1)))
        assert np.all(
            np.isclose(rs.variance, np.var(v_stack, axis=0, ddof=1)))
        with pytest.raises(Exception):
            rs.p_values

    # .....................................
    def test_matrixs_with_p_values_absolute_values(self):
        """Tests the RunningStats class using Matrix objects with p-values."""
        obs = Matrix(np.random.randint(0, 10, size=(10, 10)))
        vals = []
        rs = RunningStats(
            observed=obs, compare_fn=compare_absolute_values)
        assert np.all(rs.variance == np.zeros((10, 10)))
        for _ in range(5):
            vals.append(Matrix(np.random.randint(-10, 10, size=(10, 10))))
        rs.push(vals)
        for _ in range(5):
            test_val = Matrix(np.random.randint(-10, 10, size=(10, 10)))
            rs.push(test_val)
            vals.append(test_val)
        v_stack = np.array([v for v in vals])
        assert np.all(np.isclose(rs.mean, np.mean(v_stack, axis=0)))
        assert np.all(
            np.isclose(
                rs.standard_deviation, np.std(v_stack, axis=0, ddof=1)))
        assert np.all(
            np.isclose(rs.variance, np.var(v_stack, axis=0, ddof=1)))
        num_greater = Matrix(np.zeros((10, 10)))
        for i in range(v_stack.shape[0]):
            num_greater += np.abs(v_stack[i, ...]) > np.abs(obs)
        assert np.all(rs.p_values == num_greater.astype(float) / len(vals))

    # .....................................
    def test_matrix_with_p_values_signed(self):
        """Tests the RunningStats class using Matrix objects with p-values."""
        obs = Matrix(np.random.randint(0, 10, size=(10, 10)))
        vals = []
        rs = RunningStats(observed=obs, compare_fn=compare_signed_values)
        assert np.all(rs.variance == np.zeros((10, 10)))
        for _ in range(5):
            vals.append(Matrix(np.random.randint(-10, 10, size=(10, 10))))
        rs.push(vals)
        for _ in range(5):
            test_val = Matrix(np.random.randint(-10, 10, size=(10, 10)))
            rs.push(test_val)
            vals.append(test_val)
        v_stack = np.array([v for v in vals])
        assert np.all(np.isclose(rs.mean, np.mean(v_stack, axis=0)))
        assert np.all(
            np.isclose(
                rs.standard_deviation, np.std(v_stack, axis=0, ddof=1)))
        assert np.all(
            np.isclose(rs.variance, np.var(v_stack, axis=0, ddof=1)))
        num_greater = Matrix(np.zeros((10, 10)))
        for i in range(v_stack.shape[0]):
            num_greater += v_stack[i, ...] > obs
        assert np.all(rs.p_values == num_greater.astype(float) / len(vals))

    # .....................................
    def test_numpy_arrays_with_no_p_values(self):
        """Tests the RunningStats class using numpy arrays without p-values."""
        vals = []
        rs = RunningStats()
        assert np.all(rs.variance == np.zeros((10, 10)))
        for _ in range(5):
            vals.append(np.random.randint(-10, 10, size=(10, 10)))
        rs.push(vals)
        for _ in range(5):
            test_val = np.random.randint(-10, 10, size=(10, 10))
            rs.push(test_val)
            vals.append(test_val)
        v_stack = np.array(vals)
        assert np.all(np.isclose(rs.mean, np.mean(v_stack, axis=0)))
        assert np.all(
            np.isclose(rs.standard_deviation, np.std(v_stack, axis=0, ddof=1)))
        assert np.all(np.isclose(rs.variance, np.var(v_stack, axis=0, ddof=1)))
        with pytest.raises(Exception):
            rs.p_values

    # .....................................
    def test_numpy_arrays_with_p_values_absolute_values(self):
        """Tests the RunningStats class using numpy arrays with p-values."""
        obs = np.random.randint(0, 10, size=(10, 10))
        vals = []
        rs = RunningStats(
            observed=obs, compare_fn=compare_absolute_values)
        assert np.all(rs.variance == np.zeros((10, 10)))
        for _ in range(5):
            vals.append(np.random.randint(-10, 10, size=(10, 10)))
        rs.push(vals)
        for _ in range(5):
            test_val = np.random.randint(-10, 10, size=(10, 10))
            rs.push(test_val)
            vals.append(test_val)
        v_stack = np.array(vals)
        assert np.all(np.isclose(rs.mean, np.mean(v_stack, axis=0)))
        assert np.all(
            np.isclose(rs.standard_deviation, np.std(v_stack, axis=0, ddof=1)))
        assert np.all(np.isclose(rs.variance, np.var(v_stack, axis=0, ddof=1)))
        num_greater = np.zeros((10, 10))
        for i in range(v_stack.shape[0]):
            num_greater += np.abs(v_stack[i, ...]) > np.abs(obs)
        assert np.all(rs.p_values == num_greater.astype(float) / len(vals))

    # .....................................
    def test_numpy_arrays_with_p_values_signed(self):
        """Tests the RunningStats class using numpy arrays with p-values."""
        obs = np.random.randint(0, 10, size=(10, 10))
        vals = []
        rs = RunningStats(observed=obs, compare_fn=compare_signed_values)
        assert np.all(rs.variance == np.zeros((10, 10)))
        for _ in range(5):
            vals.append(np.random.randint(-10, 10, size=(10, 10)))
        rs.push(vals)
        for _ in range(5):
            test_val = np.random.randint(-10, 10, size=(10, 10))
            rs.push(test_val)
            vals.append(test_val)
        v_stack = np.array(vals)
        assert np.all(np.isclose(rs.mean, np.mean(v_stack, axis=0)))
        assert np.all(
            np.isclose(rs.standard_deviation, np.std(v_stack, axis=0, ddof=1)))
        assert np.all(np.isclose(rs.variance, np.var(v_stack, axis=0, ddof=1)))
        num_greater = np.zeros((10, 10))
        for i in range(v_stack.shape[0]):
            num_greater += v_stack[i, ...] > obs
        assert np.all(rs.p_values == num_greater.astype(float) / len(vals))



"""Module containing wrangler factory tools."""
from lmpy.data_wrangling.occurrence.filters import (
    get_attribute_filter, get_bounding_box_filter,
    get_decimal_precision_filter, get_disjoint_geometries_filter,
    get_intersect_geometries_filter, get_minimum_points_filter,
    get_spatial_index_filter, get_unique_localities_filter)
from lmpy.data_wrangling.occurrence.modifiers import (
    get_accepted_name_modifier, get_attribute_modifier,
    get_common_format_modifier, get_coordinate_converter_modifier)
from lmpy.spatial import SpatialIndex


# .............................................................................
class WRANGLER_TYPES:
    """Constants class for occurrence data wrangler types."""
    # Filters
    ATTRIBUTE_FILTER = 'attribute_filter'
    BBOX_FILTER = 'bbox_filter'
    DECIMAL_PRECISION_FILTER = 'decimal_precision_filter'
    DISJOINT_GEOMETRIES_FILTER = 'disjoint_geometries_filter'
    INTERSECT_GEOMETRIES_FILTER = 'intersect_geometries_filter'
    MINIMUM_POINTS_FILTER = 'minimum_points_filter'
    SPATIAL_INDEX_FILTER = 'spatial_index_filter'
    UNIQUE_LOCALITIES_FILTER = 'unique_localities_filter'
    # Modifiers
    ACCEPTED_NAME_MODIFIER = 'accepted_name_modifier'
    ATTRIBUTE_MODIFIER = 'attribute_modifier'
    ATTRIBUTE_MAP_MODIFIER = 'attribute_map_modifier'
    CORDINATE_CONVERTER_MODIFIER = 'coordinate_converter_modifier'


# .............................................................................
def wrangler_factory(wrangler_config_list):
    """Get an occurrence data wrangler from the wrangler configuration.

    Args:
        wrangler_config_list (list of dict): A list of dictionaries of wrangler
            configuration parameters.

    Returns:
        List of Method: A list of occurrence data wranglers.
    """
    wranglers = []
    for wrangler_config in wrangler_config_list:
        wrangler_type = wrangler_config['wrangler_type'].lower()
        occurrence_wrangler = None
        if wrangler_type == WRANGLER_TYPES.ATTRIBUTE_FILTER:
            att_name = wrangler_config['attribute_name']
            list_delimiter = wrangler_config['field_delimiter']
            bad_values = wrangler_config['condition']['all']['not_in']

            # Not in list
            def all_not_in(value):  # pragma: no cover
                field_values = value.split(list_delimiter)
                return all(
                    [val.strip(']').strip('[').strip(
                        '"') not in bad_values for val in field_values])

            occurrence_wrangler = get_attribute_filter(att_name, all_not_in)
        elif wrangler_type == WRANGLER_TYPES.DECIMAL_PRECISION_FILTER:
            occurrence_wrangler = get_decimal_precision_filter(
               int(wrangler_config['decimal_precision']))
        elif wrangler_type == WRANGLER_TYPES.DISJOINT_GEOMETRIES_FILTER:
            # Get geometries (wkts)
            wkts = wrangler_config['geometry_wkts']
            occurrence_wrangler = get_disjoint_geometries_filter(wkts)
        elif wrangler_type == WRANGLER_TYPES.INTERSECT_GEOMETRIES_FILTER:
            wkts = wrangler_config['geometry_wkts']
            occurrence_wrangler = get_intersect_geometries_filter(wkts)
        elif wrangler_type == WRANGLER_TYPES.UNIQUE_LOCALITIES_FILTER:
            occurrence_wrangler = get_unique_localities_filter()
        elif wrangler_type == WRANGLER_TYPES.BBOX_FILTER:
            min_x = float(wrangler_config['min_x'])
            min_y = float(wrangler_config['min_y'])
            max_x = float(wrangler_config['max_x'])
            max_y = float(wrangler_config['max_y'])
            occurrence_wrangler = get_bounding_box_filter(
                min_x, min_y, max_x, max_y)
        elif wrangler_type == WRANGLER_TYPES.MINIMUM_POINTS_FILTER:
            occurrence_wrangler = get_minimum_points_filter(
                int(wrangler_config['minimum_points']))
        elif wrangler_type == WRANGLER_TYPES.SPATIAL_INDEX_FILTER:
            spatial_index = SpatialIndex(wrangler_config['index_file'])

            def check_hit_func(hit, check_vals):  # pragma: no cover
                for check_key, check_val in check_vals:
                    if check_key in hit:
                        if hit[check_key] == check_val:
                            return True
                return False

            def get_valid_intersections_func(species_name):  # pragma: no cover
                ret_vals = []
                if species_name in wrangler_config['species']:
                    for level, value in wrangler_config['species'][species_name]:
                        level_key = 'Level_{}_cod'.format(level)
                        if int(level) < 4:
                            level_key = level_key.upper()
                        ret_vals.append((level_key, value))
                return ret_vals

            occurrence_wrangler = get_spatial_index_filter(
                spatial_index, get_valid_intersections_func, check_hit_func)
        elif wrangler_type == WRANGLER_TYPES.ATTRIBUTE_MODIFIER:
            att_name = wrangler_config['attribute_name']
            map_dict = wrangler_config['map_values']

            def replace_func(in_value):  # pragma: no cover
                if in_value in map_dict:
                    return map_dict[in_value]
                return None

            occurrence_wrangler = get_attribute_modifier(att_name, replace_func)
        elif wrangler_type == WRANGLER_TYPES.ATTRIBUTE_MAP_MODIFIER:
            mapping = wrangler_config['attribute_mapping']
            occurrence_wrangler = get_common_format_modifier(mapping)
        elif wrangler_type == WRANGLER_TYPES.ACCEPTED_NAME_MODIFIER:
            occurrence_wrangler = get_accepted_name_modifier(
                wrangler_config['filename'])
        elif wrangler_type == WRANGLER_TYPES.CORDINATE_CONVERTER_MODIFIER:
            in_epsg = wrangler_config['in_epsg']
            out_epsg = wrangler_config['out_epsg']
            occurrence_wrangler = get_coordinate_converter_modifier(
                in_epsg, out_epsg)
        wranglers.append(occurrence_wrangler)
    return wranglers



"""Module containing occurrence data wranglers for filtering points."""
from osgeo import ogr
from lmpy.data_wrangling.occurrence.common import get_occurrence_filter, make_list
from lmpy.spatial import SpatialIndex


# .............................................................................
def get_attribute_filter(att_name, pass_condition):
    """Get a filter function for checking a point attribute.

    Args:
        att_name (str): The attribute of a point to check.
        pass_condition (Method): A function that takes a value and returns a boolean
            indicating if it passes a condition.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    # .......................
    def point_filter_func(point):
        """Get the specified attribute from the point and check it.

        Args:
            point (Point): A point object to evaluate.

        Returns:
            bool: Indicator if the point passes filtering.
        """
        return pass_condition(point.get_attribute(att_name))

    return get_occurrence_filter(point_filter_func)


# .............................................................................
def get_bounding_box_filter(min_x, min_y, max_x, max_y):
    """Get a filter function for the specified bounding box.

    Args:
        min_x (numeric): The minimum 'x' value for the bounding box.
        min_y (numeric): The minimum 'y' value for the bounding box.
        max_x (numeric): The maximum 'x' value for the bounding box.
        max_y (numeric): The maximum 'y' value for the bounding box.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    # .......................
    def bbox_filter_func(point):
        """Filter for single point.

        Args:
            point (Point): A point object to evaluate.

        Returns:
            bool: Indicator if the point passes filtering.
        """
        return min_x <= point.x <= max_x and min_y <= point.y <= max_y

    return get_occurrence_filter(bbox_filter_func)


# .............................................................................
def get_decimal_precision_filter(decimal_places):
    """Get a filter function to remove points with inadequate precision.

    Args:
        decimal_places (int): Only keep points with at least this many decimal
            places of precision.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    # .......................
    def decimal_precision_filter_func(point):
        """Filter for single point.

        Args:
            point (Point): A point object to evaluate.

        Returns:
            bool: Indicator if the point passes filtering.
        """
        lat_str = str(point.y)
        lon_str = str(point.x)
        try:
            lat_decimals = len(lat_str) - lat_str.index('.') - 1
            lon_decimals = len(lon_str) - lon_str.index('.') - 1
        except ValueError:
            # TODO: Handle numbers with 'e' example: 1e-05
            return False
        return min([lat_decimals, lon_decimals]) >= decimal_places

    return get_occurrence_filter(decimal_precision_filter_func)


# .............................................................................
def get_disjoint_geometries_filter(geometry_wkts):
    """Get a filter function for finding disjoint geometries.

    Args:
        geometry_wkts (list of str): A list of geometry Well-Known Text strings
            to check for intersection.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    geom_index = SpatialIndex()
    i = 0
    for wkt in geometry_wkts:
        geom_index.add_feature(i, wkt, {"feature_id": i})
        i += 1

    # .......................
    def disjoint_geometry_filter(point):
        """Disjoint geometry filter function.

        Args:
            point (Point): A point object to evaluate.

        Returns:
            bool: Indicator if the point passes filtering.
        """
        hits = geom_index.search(point.x, point.y)
        return not bool(hits)
    return get_occurrence_filter(disjoint_geometry_filter)


# .............................................................................
def get_intersect_geometries_filter(geometry_wkts):
    """Get a filter function for intersecting the provided geometries.

    Args:
        geometry_wkts (list of str): A list of geometry Well-Known Text strings
            to check for intersection.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    geometries = []
    for wkt in geometry_wkts:
        geometries.append(ogr.CreateGeometryFromWkt(wkt))

    # .......................
    def intersect_geometry_filter(point):
        """Intersect geometry filter function.

        Args:
            point (Point): A point object to evaluate.

        Returns:
            bool: Indicator if the point passes filtering.
        """
        point_geometry = ogr.Geometry(ogr.wkbPoint)
        point_geometry.AddPoint(point.x, point.y)
        return any(
            [not geom.Intersection(
                point_geometry).IsEmpty() for geom in geometries])

    return get_occurrence_filter(intersect_geometry_filter)


# .............................................................................
def get_minimum_points_filter(minimum_count):
    """Get a filter that returns all points if a minimum number, else none.

    Args:
        minimum_count (int): The minimum number of points in order to keep.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    # .......................
    def min_points_filter(points):
        """Check that there are at least a minimum number of points.

        Args:
            points (list of Point): A list of points.

        Returns:
            list: All or nothing list of points sent to the function.
        """
        if len(make_list(points)) >= minimum_count:
            return points
        return []

    return min_points_filter


# .............................................................................
def get_spatial_index_filter(
    spatial_index,
    get_species_intersections_func,
    check_hit_func
):
    """Get a filter that uses a spatial index and logic to intersect hits.

    Args:
        spatial_index (str or SpatialIndex): A spatial index or file path for
            one to load.
        get_species_intersections_func (method): A method for retrieving valid
            intersections for a species point.
        check_hit_func (method): A method for checking a hit against the valid
            hits for a species point.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    # Load spatial index
    if not isinstance(spatial_index, SpatialIndex):
        spatial_index = SpatialIndex(spatial_index)
    # Store valid intersections for each encountered species
    valid_intersections = {}

    # .......................
    def spatial_index_point_filter(point):
        """Check a point against the spatial index.

        Args:
            point (Point): A point object to evaluate.

        Returns:
            bool: Indicator if the point passes filtering.
        """
        if point.species_name in valid_intersections:
            check_intersections = valid_intersections[point.species_name]
        else:
            check_intersections = get_species_intersections_func(
                point.species_name)
            valid_intersections[point.species_name] = check_intersections
        if check_intersections is None or len(check_intersections) == 0:
            return True
        for hit in spatial_index.search(point.x, point.y).values():
            if check_hit_func(hit, check_intersections):
                return True
        return False

    return get_occurrence_filter(spatial_index_point_filter)


# .............................................................................
def get_unique_localities_filter():
    """Get a filter that only retains points with unique localities.

    Note:
        Retains the first point with an unseen locality, drops the rest.

    Returns:
        Method: A function that takes a list of points as input and returns a list of
            points that pass that filter function.
    """
    unique_localities = []

    # .......................
    def unique_localities_filter(point):
        """Unique localities filter function.

        Args:
            point (Point): A point object to evaluate.

        Returns:
            bool: Indicator if the point passes filtering.
        """
        test_val = (point.x, point.y)
        if test_val in unique_localities:
            return False
        unique_localities.append(test_val)
        return True

    return get_occurrence_filter(unique_localities_filter)
